/*****************************************************************************
			=-= P r o j e c t =-=
	 Mining Single-Dimensional Boolean Association Rules from
	 ~~~~~~~~~~~~~~~Transactional DbaseS~~~~~~~~~~~~~~~~~~~~~
			~~~~~~~~~~~~~~~~~~~~
Ver 2.2
	16/04/2002
Update Spec:
	On the contrary to v1.0, this does not load the	transaction
	table in memory. It is scanned each time during candidate
	itemset generation from	the disk thus saving memory.
	Optimized has_infrequent_subsets

Features:
> Dynamic and on demand memory allocation
	Complete Linked list organization
> Uses Apriori Alogithm in general and Apriori property in specific
  to prune sets

Parameters:
1. Transaction Database
2. Minimum Support Threshold

TODOS:
Work out 'Transaction Database Partitioning' to further reduce the overall
time required to find candidate itemsets.
*****************************************************************************/
# include <alloc.h>
# include <stdlib.h>
# include <stdio.h>
# include <math.h>
# include <assert.h>
# include <conio.h>
# include <string.h>
# include <ctype.h>
# include "apritrn.h"

extern int debugmode = OFF ; //ON/OFF
int main ( int argc , char *argv[] )
{
	FILE *fp ;
	unsigned int iter = 1 ;

	clrscr ( ) ;
	if ( argc != 3 )
	{
		printf ( "\nUsage: rowl <infile> <min_sup>\n" ) ;
		exit ( 1 ) ;
	}

	if ( !atol ( argv[2] ) )
		if ( strlen ( argv[2] ) == 1 && argv[2][0] == '0' )
			;
		else
			error_exit ( "sup_cnt format" ) ;
	SHOWMEM
	fp = fopen ( argv[1] , "r" ) ;
	if ( !fp )
		error_exit ( "input file open" ) ;

	//Create C1
	clrscr ( ) ;
	CreateList ( fp ) ;
	fclose ( fp ) ;

	APRI_DEBUG (( "\nDisplaying C1:" )) ;
	DisplayList ( firstC ) ;
	SHOWMEM
	//Create L1
	currL = find_frequent_1 ( atol ( argv[2] ) ) ;

	printf ( "\nGenerated C1 & L1 ..." ) ;
	// Free C1
	while ( firstC )
	{
		struct list *tmp ;
		tmp = firstC ;
		firstC = firstC -> next_ptr ;
		free ( tmp ) ;
	}

	DisplayFreqList ( currL , "L" ) ;

	// while there are still some frequent list items
	while ( currL )
	{
		struct cf *tc ;
		fp = fopen ( argv[1] , "r" ) ;
		if ( !fp )
			error_exit ( "file open" ) ;
		apriori_gen ( currL ) ;
		if ( currC -> startcf )
		{ // currC has got some itemids
			unsigned long tid = 0 ;
			++iter ;
			printf ( "\n---------------------------------------------" ) ;
			printf ( "\nGenerating C%u & L%u...." , iter , iter ) ;
			printf ( "\nPress <ESC> to abort...\n" ) ;
			while ( GetNextTransaction ( fp ) && !kbhit() )
			{ // Got next transaction from file and read it into trn
				printf ( "\rScanning transaction : %lu" , ++tid ) ;
				tc = currC -> startcf ;
				while ( tc )
				{
					if ( has_candidate ( tc -> iid , trn.iid , trn.ino , currC -> setlen ) )
						tc -> sup_cnt++ ;
					tc = tc -> next ;
				}
				free ( trn.iid ) ;
			}
			if ( kbhit() )
				if ( getch() == ESC )
				{
					printf ( "\nUser Request:Terminating..." ) ;
					exit ( 0 ) ;
				}
		}
		else
		{
			freecflist ( currC ) ;
			currC = NULL ;
		}
		freecflist ( currL ) ;

		//stopped here
		if ( currC )
		{
			SupcntPrune ( atol ( argv[2] ) ) ;
			DisplayFreqList ( currC , "C" ) ;
			if ( currL )
			{
				//(over)write frequent itemsets file
				WriteLk2file ( currL ) ;
				// Display frequent itemsets file
				DisplayFreqList ( currL , "L" ) ;
			}
		}
		else
			currL = NULL ; //That's it with frequent itemsets
		fclose ( fp ) ;
	}
	SHOWMEM
	return 0 ;
}

int WriteLk2file ( struct cflist *tmp )
{
	struct cf *tmpcf ;
	unsigned int setlen ;
	FILE *fp ;

	printf ( "\n(over)writing Lk to <%s>..." , LKFILE ) ;
	if ( ! ( tmp -> startcf ) )
		return 0 ;

	tmpcf 	= tmp -> startcf ;
	setlen 	= tmp -> setlen ;

	if ( !( fp = fopen ( LKFILE , "w" ) ) )
	{
		printf ( "\nUnable to write Lk to <%s>" , LKFILE ) ;
		return 0 ;
	}

	while ( tmpcf )
	{
		register int i ;
		for ( i = 0 ; i < setlen ; i++ )
			fprintf ( fp , " %d" , tmpcf -> iid[i] ) ;
		fprintf ( fp , "\n" ) ;
		tmpcf = tmpcf -> next ;
	}
	fclose ( fp ) ;
	return 0 ;
}

// Read next transaction by storing it into DS trn
int GetNextTransaction ( FILE *fp )
{
	// Above value should be incremented for bigger files like connect4.dat
	char rowstr[MAXROW + 1] ;
	unsigned int len ;
	char *tptr , *ptr ;
	unsigned int flag = 0 , itemnos = 0 ;
	unsigned int itemids[MAXITM] ;
	// Read a line and process
	if ( !fgets ( rowstr , MAXROW + 1 , fp ) )
		return NULL ;
	len = strlen ( rowstr ) ;
	if ( rowstr[len - 1] == '\n' )
		rowstr[len - 1] = ' ' ; // Convert last \n to space
	else
	{
		rowstr[len] = ' ' ;
		rowstr[len + 1] = '\0' ;
	}

	tptr = ptr = rowstr ;
	while ( *ptr )
	{
		if ( isdigit ( *ptr ) )
			ptr++ , flag = 1 ;
		else
		{
			if ( isspace ( *ptr ) )
			{
				if ( flag )
				{
					*ptr = '\0' ;
					itemids[itemnos++] = atoi ( tptr ) ;
					if ( itemnos >= MAXITM )
						error_exit ( "itemnos exceeds MAXITEMS" ) ;
				}
				ptr++ ;
				tptr = ptr ;
				flag = 0 ;
			}
			else
				error_exit ( "input format" ) ;
		}
	}
	if ( itemnos )
	{
		register int i ;
		trn.iid = ( unsigned int * ) malloc ( sizeof ( unsigned int ) * itemnos ) ;
		trn.ino = itemnos ;
		for ( i = 0 ; i < itemnos ; i++ )
			trn.iid[i] = itemids[i] ;
	}
	return TRUE ;
}

// Create new Lk from Ck by pruning nodes from Ck using min_sup
int SupcntPrune ( unsigned long min_supcnt )
{
	struct cflist *tmpL ;
	struct cf *c , *tc ;
	int flag = 0 ;

	tmpL = ( struct cflist * ) calloc ( 1 , sizeof ( struct cflist ) ) ;
	if ( !tmpL )
		error_exit ( "mem alloc" ) ;
	tmpL -> setlen = currC -> setlen ;
	tc = currC -> startcf ;
	while ( tc )
	{
		if ( tc -> sup_cnt < min_supcnt )
			; // Node Prunned
		else
		{       // Node should be created and copied
			register int i ;
			flag = 1 ;
			c = ( struct cf * ) calloc ( 1 , sizeof ( struct cf ) ) ;
			if ( !c )
				error_exit ( "mem alloc" ) ;
			c -> iid = ( unsigned int * ) malloc ( sizeof ( unsigned int ) * currC -> setlen ) ;
			c -> sup_cnt = tc -> sup_cnt ;
			c -> next = NULL ;
			for ( i = 0 ; i < currC -> setlen ; i++ )
				c -> iid[i] = tc -> iid[i] ;
			if ( ! ( tmpL -> startcf ) )
				tmpL -> startcf =
				tmpL -> endcf =
				c ;
			else
			{
				tmpL -> endcf -> next = c ;
				tmpL -> endcf = c ;
			}
		}
		tc = tc -> next ;
	}
	if ( !flag )
	{
		free ( tmpL ) ;
		currL = NULL ;
	}
	else
	{
		if ( tmpL -> startcf )
			currL = tmpL ;
	}
	return 0 ;
}

// Free DS cflist
int freecflist ( struct cflist *tmpL )
{
	struct cf *tmp ;
	tmp = tmpL -> startcf ;

	while ( tmp )
	{
		struct cf *tmp1 ;
		tmp1 = tmp ;
		tmp = tmp -> next ;
		free ( tmp1 -> iid ) ; // allocated items should be freed
		free ( tmp1 ) ;
	}
	free ( tmpL ) ;
	return 0 ;
}

// Sees whether the candydate is present in current transaction
int has_candidate ( tciid , tmpiid , trnino , tcsetlen )
unsigned int *tciid , *tmpiid ;
unsigned int trnino , tcsetlen ;
{
	unsigned int i , j , flag = 0 ;

	for ( i = 0 ; i < trnino ; i++ )
	{
		for ( j = 0 ; j < tcsetlen ; j++ )
		{
			if ( tciid[j] == tmpiid[i] )
			{
				flag++ ;
				break ;
			}
		}
	}
	if ( flag == tcsetlen )
		return TRUE ;
	return FALSE ;
}

// Std apriori algo
int apriori_gen ( struct cflist *tmpL )
{
	struct cf *l1 , *l2 ; // Make two pointers to the list
	struct cflist *newL ;
	struct cf *c ;

	l1 = tmpL -> startcf ; //2 pointers pointing to same list
	newL = ( struct cflist * ) calloc ( 1 , sizeof ( struct cflist ) ) ;
	if ( !newL )
		error_exit ( "mem alloc" ) ;
	newL -> setlen = ( tmpL -> setlen ) + 1 ;
	while ( l1 )
	{
		l2 = l1 -> next ;
		while ( l2 )
		{
			register int flag = 0 , i ;
			for ( i = 0 ; i < tmpL -> setlen ; i++ )
			{
				if ( ( ( tmpL -> setlen ) - 1 ) == i )
				{
					if ( l1 -> iid[i] < l2 -> iid[i] )
						flag = 1 ;
				}
				else
				{
					if ( l1 -> iid[i] == l2 -> iid[i] )
						;
					else
					{
						flag = 0 ;
						break ;
					}
				}
			}
			if ( flag )
			{
				c = ( struct cf * ) malloc ( sizeof ( struct cf ) ) ;
				if ( !c )
					error_exit ( "mem alloc" ) ;
				c -> iid = ( unsigned int * ) malloc ( newL -> setlen * sizeof ( unsigned int ) ) ;
				if ( ! ( c -> iid ) )
					error_exit ( "mem alloc" ) ;
				c -> sup_cnt = 0 ;
				join ( l1 , l2 , newL , c ) ;
				if ( has_infrequent_subset ( c , tmpL ) )
				{
					free ( c -> iid ) ;
					free ( c ) ;
				}
				else
				{ // add c to Ck
					if ( ! ( newL -> startcf ) )
					{
						newL -> startcf =
						newL -> endcf   =
						c ;
					}
					else
					{
						newL -> endcf -> next = c ;
						newL -> endcf = c ;
					}
				}
			}
			l2 = l2 -> next ;
		}
		l1 = l1 -> next ;
	}
	currC = newL ;
	return 0 ;
}

/* faced most difficulties while designing this routine
   reverse engineered from JAVA code
   routine deals with the problem of combinatorials
   thanksto : sujatha & anagha */
int has_infrequent_subset ( struct cf *c , struct cflist *tmpL )
{
	unsigned int *st , *ptr1 , len , i , j , *stpt1 ;
	len = tmpL -> setlen ;
	// allocate space for tmp subset
	stpt1 = ( unsigned int * ) malloc ( sizeof ( unsigned int ) * len ) ;
	for ( i = 0 ; i < len - 1 ; i++ )
	{
		st = c -> iid ;
		ptr1 = stpt1 ;
		for ( j = 0 ; j < i ; j++ )
			*ptr1++ = *st++ ;
		st++ ;
		for ( j = i + 1 ; j < len + 1 ; j++ )
			*ptr1++ = *st++ ;
		if ( !Chek4SubsetinTrn ( stpt1 , tmpL ) )
		{
			free ( stpt1 ) ;
			return TRUE ;
		}
	}
	free ( stpt1 ) ;
	return FALSE ;
}

// Checks whether itemset is present in earlier Frequent itemset list
int Chek4SubsetinTrn ( unsigned int *ptr , struct cflist *tmpL )
{
	struct cf *scf ;
	unsigned int *tmp ;

	scf = tmpL -> startcf ;
	while ( scf )
	{
		tmp = scf -> iid ;
		if ( !memcmp ( ptr , tmp , sizeof ( unsigned int ) * tmpL -> setlen - 1 ) )
			return TRUE ;
		scf = scf -> next ;
	}
	return FALSE ;
}

// 1><1 operation
int join ( struct cf *l1 , struct cf *l2 , struct cflist *newL , struct cf *newC )
{
	register int i ;

	for ( i = 0 ; i < ( newL -> setlen ) - 1 ; i++ )
		newC -> iid[i] = l1 -> iid[i] ;
	newC -> iid[i] = l2 -> iid[i-1] ;
	newC -> next = NULL ;
	return 0 ;
}

struct cflist * find_frequent_1 ( unsigned long min_supcnt )
{
	struct list *tmp ;
	struct cflist *newcflist ;
	struct cf *newcf ;

	newcflist = ( struct cflist * ) calloc ( 1 , sizeof ( struct cflist ) ) ;
	newcflist -> setlen = 1 ;
	tmp = firstC ;
	if ( !tmp )
		error_exit ( "no frequent items" ) ;
	while ( tmp )
	{
		if ( tmp -> sup_cnt >= min_supcnt )
		{
			newcf = ( struct cf * ) malloc ( sizeof ( struct cf ) ) ;
			assert ( newcf ) ;
			newcf -> iid = ( unsigned int * ) malloc ( ( newcflist -> setlen ) * ( sizeof ( unsigned int ) ) ) ;
			* ( newcf -> iid ) = tmp -> item_id ;
			newcf -> sup_cnt = tmp -> sup_cnt ;
			newcf -> next = NULL ;
			if ( ! ( newcflist -> startcf ) )// First L itemset
				newcflist -> startcf =
				newcflist -> endcf   =
				newcf ;
			else
			{
				newcflist -> endcf -> next = newcf ;
				newcflist -> endcf = newcf ;
			}
		}
		tmp = tmp -> next_ptr ;
	}
	return newcflist ;
}

int Create_Node ( unsigned int value )
{
	struct list *tmp , *prevtmp , *newnode ;

	newnode = ( struct list * ) malloc ( sizeof ( struct list ) ) ;
	if ( !newnode )
		error_exit ( "mem alloc" ) ;
	newnode -> item_id = value ;
	newnode -> sup_cnt = 1 ;
	newnode -> next_ptr = NULL ;

	if ( !firstC )
		LastC = firstC = newnode ;
	else
	{
		prevtmp = tmp = firstC ;
		while ( tmp )
		{
			if ( tmp -> item_id > value )
			{	// Append as first or intermediate node
				if ( tmp == prevtmp )
				{
					newnode -> next_ptr = tmp ;
					firstC = newnode ;
				}
				else
				{
					prevtmp -> next_ptr = newnode ;
					newnode -> next_ptr = tmp ;
				}
				break ;
			}
			if ( tmp -> item_id == value )
			{	// Duplicate entry so, freeing tied mem
				tmp -> sup_cnt++ ;
				free ( newnode ) ;
				break ;
			}
			prevtmp = tmp ;
			tmp = tmp -> next_ptr ;
		}
		if ( !tmp )
		{	// Last Node
			LastC -> next_ptr = newnode ;
			LastC = newnode ;
		}
	}
	return 0 ;
}

int CreateList ( FILE *fp )
{
	# define MAXROWSIZE 150
	// Above value should be incremented for bigger files like connect4.dat
	char rowstr[MAXROWSIZE] ;
	unsigned int len ;
	unsigned long tid = 0 ;
	// Read a line and process
	while ( fgets ( rowstr , MAXROWSIZE , fp ) )
	{
		len = strlen ( rowstr ) ;
		if ( len == MAXROWSIZE - 1 )
			error_exit ( "MAXROWSIZE limit" ) ;
		if ( rowstr[len - 1] == '\n' )
			rowstr[len - 1] = ' ' ; // Convert last \n to space
		Process_Row ( rowstr ) ;
		printf ( "\rProcessing transaction : %lu" , ++tid ) ;
	}
	return 0 ;
}

int Process_Row ( char *ptr )
{
	char *tptr ;
	unsigned int flag = 0 ;

	tptr = ptr ;
	while ( *ptr )
	{
		if ( isdigit ( *ptr ) )
			ptr++ , flag = 1 ;
		else
		{
			if ( isspace ( *ptr ) )
			{
				if ( flag )
				{
					*ptr = '\0' ;
					Create_Node ( atoi ( tptr ) ) ;
				}
				ptr++ ;
				tptr = ptr ;
				flag = 0 ;
			}
			else
				error_exit ( "input format" ) ;
		}
	}
	return 0 ;
}

// dISPLAY DS cflist
int DisplayFreqList ( struct cflist *tmp , char *ptr )
{
	int j = 1 ;
	struct cf *tmpcf ;

	tmpcf = tmp -> startcf ;
	if ( tmpcf )
		APRI_DEBUG (( "\n%s%d:" , ptr , tmp -> setlen )) ;
	else
		return 0 ;

	APRI_DEBUG (( "\n%-10s%3sItemSet" , "SupCnt" , " " )) ;
	while ( tmpcf )
	{
		int i ;
		APRI_DEBUG (( "\n" )) ;
		APRI_DEBUG (( "%-10lu{" , tmpcf -> sup_cnt )) ;
		for ( i = 0 ; i < tmp -> setlen ; i++ )
			APRI_DEBUG (( "%5d" , tmpcf -> iid[i] )) ;
		APRI_DEBUG (( "%5s" , "}" )) ;
		tmpcf = tmpcf -> next ;
		if ( j++ % 20 == 0 )
		{
			APRI_DEBUG (( "\nPress a key to continue..." )) ;
			if ( debugmode ) getch ( ) ;
		}
	}
	SHOWMEM
	if ( debugmode ) getch ( ) ;
	return 0 ;
}

int DisplayList ( struct list *tmpfirst )
{
	int i = 1 ;
	struct list *tmp ;
	tmp = tmpfirst ;
	while ( tmp )
	{
		if ( i++ % 20 == 0 )
		{
			APRI_DEBUG (( "\nPress a key to continue..." )) ;
			if ( debugmode ) getch ( ) ;
		}
		APRI_DEBUG (( "\n%u\t\t%lu" , tmp -> item_id , tmp -> sup_cnt )) ;
		tmp = tmp -> next_ptr ;
	}
	return 0 ;
}

// Display failure reason and Quit Abnormally
void error_exit ( char *ptr )
{
	printf ( "\n%s failure...\nTerminating Abnormally.\n" , ptr ) ;
	exit ( 1 ) ;
}

