/*****************************************************************************
Generating Association Rules from Frequent Itemset(s)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Version:0.6
Date:	28/03/2002

Basic Rule:
confidence(A=>B) = P(B/A) = support_count(AUB)
			    ------------------
			    support_count(A)

Features:
> makes use of state machine to read input file and tokenize
> dynamic on demand memory allocation
> portable code ( tested on Red Hat Linux 7.2 , Win98 )
> extensive error checking

Parameters:
1. Transaction Database file
2. Frequent Itemset(s) list file
3. Maximum Confidence Threshold
*****************************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <mem.h>
#include <dos.h>

#define TRUE 	1
#define FALSE 	0
#define ON	1
#define OFF	0
#define MAXIDS 	50
#define MAXSTR 	200
#define GENRUL	"associat.rul"
#define GEN_DEBUG(x) if(debugmode) printf x

void error_exit 	( char * ) ;
int main 		( int , char *[] ) ;
int StoreAssocRules 	( void ) ;
int PrintAssocRules 	( unsigned int * , unsigned int ) ;
int CountOccurence 	( unsigned int * , unsigned int ) ;
int ReadNextLk 		( FILE * ) ;
int findsubset 		( unsigned int * , unsigned int ) ;
int GetNextTransaction 	( FILE * ) ;
int has_candidate 	( unsigned int * , unsigned int * ,
			unsigned int , unsigned int ) ;


int 		debugmode = ON ;// ON/OFF Debug mode
unsigned int 	*iid ; 		// current frequent itemset
unsigned int 	setlen ;   	// no. of items in frequent itemset

struct Trn
{
	unsigned long 	tid ;
	unsigned int 	ino ;
	unsigned int 	*iid ; /*Elaborates clause 2 of features mentioned above
				stores just base address of list of item ids */
} trn ;

struct Rules
{
	unsigned int 	*siid ;
	unsigned int 	*diid ;
	unsigned int 	slen ;
	unsigned int 	dlen ;
	float 		conf ;
	struct Rules 	*next ;
} *currR , *currL ;

unsigned int sup_cntAUB , sup_cntA ;
char *tdb ;
float min_conf ;

int main ( int argc , char *argv[] )
{
	FILE *fp ;
	unsigned int lk = 0 ;

	if ( argc != 4 )
	{
		printf ( "\nUsage: genrules <trn db> <infile> <min_conf>\n" ) ;
		exit ( 1 ) ;
	}
	clrscr ( ) ;
	tdb = argv[1] ;
	/* parameter min_conf:
		can be: 0 , 50 , 50.0 , 100.5 , 100 ....
		cannot be: 0.0 , 0.00 .... */
	if ( ! ( min_conf = atof ( argv[3] ) ) )
		if ( ( argv[3][0] != '0') || ( strlen ( argv[3] ) != 1 ) )
			error_exit ( "confidence threshold read" ) ;
		else
			GEN_DEBUG (("\nMinimum confidence threshold : %3.2f" , min_conf )) ;
	fp = fopen ( argv[2] , "r" ) ;
	if ( !fp )
		error_exit ( "input file open" ) ;

	GEN_DEBUG (( "\nGenerating association rules from <%s>..." , tdb )) ;
	while ( ReadNextLk ( fp ) )
	{
		GEN_DEBUG (( "\nUsing Lk# %u" , ++lk )) ;
		sup_cntAUB = CountOccurence ( iid , setlen ) ;
		findsubset ( iid , setlen ) ;
	}
	fclose ( fp ) ;
	GEN_DEBUG (("\nWriting rules to file <%s>..." , GENRUL )) ;
	StoreAssocRules ( ) ;
	return 0 ;
}

// Store Resultant association rules
int StoreAssocRules ( void )
{
	FILE *fp ;
	struct date d ;
	struct time t ;

	fp = fopen ( GENRUL , "w" ) ;
	if ( !fp )
	{
		printf ( "\nUnable to open file <%s>" , GENRUL ) ;
		printf ( "\nDumping to stdout...\n" ) ;
		fp = stdout ;
	}

	fprintf ( fp , "=-=-= Generated Association Rules =-=-=\n" ) ;
	getdate ( &d ) ;
	fprintf ( fp , "Day : %d/%d/%d\n" , d.da_day , d.da_mon , d.da_year ) ;
	gettime ( &t ) ;
	fprintf ( fp , "Time: %2d:%02d:%02d\n" , t.ti_hour , t.ti_min , t.ti_sec ) ;
	while ( currR )
	{
		register int i ;
		for ( i = 0 ; i < currR -> slen ; i++ )
			fprintf ( fp , "%3d" , currR -> siid[i] ) ;
		fprintf ( fp , "%4c>" , '=' ) ;
		for ( i = 0 ; i < currR -> dlen ; i++ )
			fprintf ( fp , "%3d" , currR -> diid[i] ) ;
		fprintf ( fp , "\tConfidence = %3.2f\n" , currR -> conf ) ;
		currR = currR -> next ;
	}
	fclose ( fp ) ;
	return 0 ;
}

// Resultant Association Rules
int PrintAssocRules ( unsigned int *tmp , unsigned int len )
{
	struct Rules *trl ;
	register unsigned int i = 0 , j ;
	unsigned int ditem[MAXIDS] , count = 0 ;
	float conf ;

	conf = ( float ) sup_cntAUB / sup_cntA * 100 ;
	if ( conf < min_conf ) // Does not satisfy confidence threshold value
		return 0 ;
	/* fill in the diid by the items not covered in siid */
	for ( i = 0 ; i < setlen ; i++ )
	{
		int flag = 0 ;
		for ( j = 0 ; j < len ; j++ )
			if ( iid[i] == tmp[j] )
				flag = 1 ;
		if ( !flag )
			ditem[count++] = iid[i] ;
	}
	// If redundant entry
	if ( currR )
	{
		trl = currR ;
		while ( trl )
		{
			// see if all the fields siid,diid,slen,dlen match
			if ( ( len == trl -> slen ) &&
				( count == trl -> dlen ) &&
				( !memcmp ( trl -> siid , tmp , sizeof ( unsigned int ) * len ) ) &&
				( !memcmp ( trl -> diid , ditem , sizeof ( unsigned int ) * count ) ) )
					return 0 ;
			trl = trl -> next ;
		}
	}

	trl = ( struct Rules * ) malloc ( sizeof ( struct Rules ) ) ;
	if ( !trl )
		error_exit ( "mem alloc" ) ;
	// Calculate Confidence
	trl -> conf = conf ;
	trl -> siid = ( unsigned int * ) malloc ( sizeof ( unsigned int ) * len ) ;
	if ( !(trl -> siid) )
		error_exit ( "mem alloc" ) ;
	trl -> slen = len ;
	if ( !memcpy ( trl -> siid , tmp , sizeof ( unsigned int ) * len ) )
		error_exit ( "memcpy" ) ;
	trl -> diid = ( unsigned int * ) malloc ( sizeof ( unsigned int ) * count ) ;
	if ( !(trl -> diid) )
		error_exit ( "mem alloc" ) ;
	if ( !memcpy ( trl -> diid , ditem , sizeof ( unsigned int ) * count ) )
		error_exit ( "memcpy" ) ;
	trl -> dlen = count ;
	trl -> next = NULL ;
	if ( !currR )
		currR = currL = trl ;
	else
	{
		currL -> next = trl ;
		currL = trl ;
	}
	return 0 ;
}

// Count how many times specified itemset occurs in tdb
int CountOccurence ( unsigned int *tmp , unsigned int len )
{
	FILE *fp ;
	unsigned long tid = 0 ;
	unsigned int count = 0 ;

	fp = fopen ( tdb , "r" ) ;
	if ( !fp )
		error_exit ( "transaction db" ) ;

	GEN_DEBUG (( "\n" )) ;
	while ( GetNextTransaction ( fp ) && !kbhit() )
	{ // Got next transaction from file and read it into trn
		GEN_DEBUG (( "\rScanning transaction %lu" , ++tid )) ;
		if ( has_candidate ( tmp , trn.iid , trn.ino , len ) )
			count++ ;
		free ( trn.iid ) ;
	}
	if ( kbhit ( ) )
		if ( getch ( ) == 27 )
			printf ( "\nUser Request:Termination...\n" ) , exit ( 0 ) ;
	fclose ( fp ) ;
	return count ;
}

/* Read next frequent itemset from tdb and store it in trn
	Extensive error checking is not required since file read here was
	generated through a program (user cannot directly create this file).
	Assertions tackle error checking(if any) */
int ReadNextLk ( FILE *fp )
{
	char rowstr[MAXSTR] , *ptr , *tptr ;
	unsigned int /*curr inp*/cin , /*curr state*/cst , flag , len ;
	unsigned int item[MAXIDS] ;
	// state machine function
	unsigned int q[3][3] =
	{
		/* isspace , isdigit , '\0' */
		{ 0 , 1 , 2 } , /* current state is space */
		{ 0 , 1 , 2 } , /* current state is digit */
		{ 2 , 2 , 2 }  /* halt state */
	} ;

	if ( !fgets ( rowstr , MAXSTR , fp ) )
		return 0 ;

	len = strlen ( rowstr ) ;
	if ( rowstr[len - 1] == '\n' )
		rowstr[len - 1] = ' ' ; // Convert last \n to space
	tptr = ptr = rowstr ;
	cin = cst = flag = len = 0 ; // initial state & input , flag & #iid
	while ( cst != 2 ) // until final state is reached
	{
		if ( isdigit ( *ptr ) )
			cin = 1 ;
		else
			if ( isspace ( *ptr ) )
				cin = 0 ;
			else
				if ( *ptr == '\0' )
					cin = 2 ;
				else
					error_exit ( "input format" ) ;
		//Change current state = q[curr state][curr input]
		cst = q[cst][cin] ;
		switch ( cst )
		{
			register int i ;
			case 0 :
				cin = 0 ;
				if ( flag )
				{
					flag = 0 ;
					*ptr++ = '\0' ;
					item[len++] = atoi ( tptr ) ;
					tptr = ptr ;
				}
				else
					ptr++ ;
				break ;
			case 1 :
				flag = 1 ;
				ptr++ ;
				break ;
			case 2 :
				setlen = len ;
				assert ( setlen ) ;
				iid = ( unsigned int * ) malloc ( sizeof ( unsigned int ) * setlen ) ;
				if ( !iid )
					error_exit ( "mem alloc" ) ;
				for ( i = 0 ; i < setlen ; i++ )
					iid[i] = item[i] ;
				break ;
		}
	}
	return TRUE ;
}

/* finds subset recursively
   eg, 1 2 3 = 1 2 , 1 3 , 2 3 , 1 , 2 , 3
   most optimized & enjoyed routine (uses recursion) */
int findsubset ( unsigned int *tiid , unsigned int len )
{
	unsigned int *st , *ptr1 , i , j , *stpt1 ;
	stpt1 = ( unsigned int * ) malloc ( sizeof ( unsigned int ) * len ) ;
	for ( i = 0 ; i < len ; i++ )
	{
		st = tiid ;
		ptr1 = stpt1 ;
		for ( j = 0 ; j < i ; j++ )
			*ptr1++ = *st++ ;
		st++ ;
		for ( j = i + 1 ; j < len + 1 ; j++ )
			*ptr1++ = *st++ ;
		sup_cntA = CountOccurence ( stpt1 , len - 1 ) ;
		PrintAssocRules ( stpt1 , len - 1 ) ;
		if ( len - 2 )
			//recurse using new subset partition
			findsubset ( stpt1 , len - 1 ) ;
	}
	free ( stpt1 ) ;
	return 0 ;
}

// Read next transaction by storing it into DS trn
int GetNextTransaction ( FILE *fp )
{
	# define MAXROWSIZE 200
	// Above value should be incremented for bigger files like connect.dat
	char rowstr[MAXROWSIZE + 1] ;
	unsigned int len ;
	char *tptr , *ptr ;
	unsigned int flag = 0 , itemnos = 0 ;
	# define MAXITEMS 150
	unsigned int itemids[MAXITEMS] ;
	// Read a line and process
	if ( !fgets ( rowstr , MAXROWSIZE , fp ) )
		return NULL ;
	len = strlen ( rowstr ) ;
	if ( len == MAXROWSIZE - 1 )
		error_exit ( "MAXROWSIZE limit" ) ;
	if ( rowstr[len - 1] == '\n' )
		rowstr[len - 1] = ' ' ; // Convert last \n to space

	tptr = ptr = rowstr ;
	while ( *ptr )
	{
		if ( isdigit ( *ptr ) )
			ptr++ , flag = 1 ;
		else
		{
			if ( isspace ( *ptr ) )
			{
				if ( flag )
				{
					*ptr = '\0' ;
					itemids[itemnos++] = atoi ( tptr ) ;
					if ( itemnos >= MAXITEMS )
						error_exit ( "MAXITEMS limit" ) ;
				}
				ptr++ ;
				tptr = ptr ;
				flag = 0 ;
			}
			else
				error_exit ( "input format" ) ;
		}
	}
	if ( itemnos )
	{
		register int i ;
		trn.iid = ( unsigned int * ) malloc ( sizeof ( unsigned int ) * itemnos ) ;
		trn.ino = itemnos ;
		for ( i = 0 ; i < itemnos ; i++ )
			trn.iid[i] = itemids[i] ;
	}
	return TRUE ;
}

// Sees whether the candydate is present in current transaction
int has_candidate ( tciid , tmpiid , trnino , tcsetlen )
unsigned int *tciid , *tmpiid ;
unsigned int trnino , tcsetlen ;
{
	unsigned int i , j , flag = 0 ;

	for ( i = 0 ; i < trnino ; i++ )
	{
		for ( j = 0 ; j < tcsetlen ; j++ )
		{
			if ( tciid[j] == tmpiid[i] )
			{
				flag++ ;
				break ;
			}
		}
	}
	if ( flag == tcsetlen )
		return TRUE ;
	return FALSE ;
}

void error_exit ( char *ptr )
{
	printf ( "\nAbnormal termination: %s failure." , ptr ) ;
	exit ( 1 ) ;
}
